@@ -9,7 +9,7 @@
 -include("megaco_text_mini_parser.hrl").
 
 
--file("/net/isildur/ldisk/daily_build/21_prebuild_opu_o.2018-12-10_21/otp_src_21/bootstrap/lib/parsetools/include/yeccpre.hrl", 0).
+-file("/buildroot/otp/bootstrap/lib/parsetools/include/yeccpre.hrl", 0).
 %%
 %% %CopyrightBegin%
 %%
